"""
Lookup for HTTPS queries
"""

from http.client import HTTPResponse, HTTPSConnection, socket  # type: ignore
from ssl import _create_unverified_context
from time import sleep, time
from typing import Any, ClassVar, Dict, Optional
from urllib.parse import urlencode

from ..bibtex.normalize import normalize_url
from ..utils.constants import CONNECTION_TIMEOUT, MIN_QUERY_DELAY, USER_AGENT
from ..utils.logger import Hint, logger
from ..utils.safe_json import JSONType
from .abstract_base import AbstractDataLookup, Data, Input, Output

DNS_Fail_Hint = Hint("check your internet connection or DNS server")
SSL_Fail_Hint = Hint(
    "run 'pip install --upgrade certifi' to update certificates\n"
    "or run btac with the -S / --ignore-ssl flag."
)
TIMEOUT_Hint = Hint("you can increase timeout with -t / --timeout option.")


class HTTPSLookup(AbstractDataLookup[Input, Output]):
    """Abstract class to wrap https queries:
    Initialized with the entry to query info about

    Defines:
    - lookup : Self -> Optional[bytes] - performs the queries and returns raw data
    - query : Self -> Optional[BibtexEntry] - performs single query, calls lookup and handle_output

    - domain: str = "localhost" - the domain name e.g. api.crossref.org
    - host : Optional[str] = None - host when different from domain
    - path : str = "/" - the path component of the URL
    - params : Dict[str, str] = {} - parameters to add to the URL

    - request : str = "GET" - https request type
    - default_headers : Dict[str, str] = ... default http header
    - headers : Dict[str, str] = {} - headers to add, overrite default_headers

    all of these have associated methods get_XX : Self -> Type[XX] that can be overridden
    for finer behavior control

    Virtual methods and attributes:
    - handle_output : Self, bytes -> Optional[Result] - parses output into useful result
    """

    domain: str = "localhost"
    host: Optional[str] = None
    path: str = "/"
    params: Dict[str, str] = {}

    ignore_ssl: ClassVar[bool] = False
    silent_fail: ClassVar[bool] = False

    # Safe parameters kept in urlencode
    safe: str = ""

    request: str = "GET"
    accept: str = "application/json"
    default_headers: Dict[str, str] = {
        "User-Agent": USER_AGENT,
        "Accept": accept,
    }
    headers: Dict[str, str] = {}

    connection_timeout: Optional[float] = CONNECTION_TIMEOUT

    response: Optional[HTTPResponse] = None

    _last_query_info: Dict[str, JSONType] = {}

    def get_headers(self) -> Dict[str, str]:
        """Return the headers used in an HTTPS request"""
        headers = self.default_headers.copy()
        headers["Accept"] = self.accept
        headers.update(self.headers)
        headers["Host"] = self.get_host()
        return headers

    def get_request(self) -> str:
        """Return the request method to use
        override this if not using self.request (default GET)"""
        return self.request

    def get_domain(self) -> str:
        """Return the path to connect to
        override this if not using self.domain"""
        return self.domain

    def get_host(self) -> str:
        """Return the host header
        override this if not using self.host or self.domain"""
        if self.host is not None:
            return self.host
        return self.get_domain()

    def get_path(self) -> str:
        """Return the path to connect to
        override this if not using self.path"""
        params = self.get_params()
        if params:
            return self.path + "?" + urlencode(params, safe=self.safe)
        return self.path

    def get_params(self) -> Dict[str, str]:
        """Url parameters, can use self.entry to set them
        override this if not using self.path"""
        return self.params

    def get_body(self) -> Optional[Any]:
        """Query body, can use self.entry to set them"""
        return None

    def get_data(self) -> Optional[Data]:
        """main lookup function
        returns true if the lookup succeeded in finding all info
        false otherwise"""
        domain = self.get_domain()
        request = self.get_request()
        path = self.get_path()
        headers = self.get_headers()
        url = f"https://{domain}{path}"
        self._last_query_info = dict()
        logger.debug(
            "{request} {url}",
            request=request,
            url=url,
        )
        logger.very_verbose_debug("headers: {headers}", headers=headers)
        start = time()
        try:
            if self.ignore_ssl:
                connection = HTTPSConnection(
                    domain,
                    timeout=self.connection_timeout,
                    context=_create_unverified_context(),
                )
            else:
                connection = HTTPSConnection(domain, timeout=self.connection_timeout)
            connection.request(
                request,
                path,
                self.get_body(),
                headers,
            )
            self.response = connection.getresponse()
            delay = round(time() - start, 3)
            self._last_query_info = {
                "url": url,
                "response-time": delay,
                "response-status": self.response.status,
            }
            logger.debug(
                "response: {status}{reason} in {delay}s",
                status=self.response.status,
                reason=" " + self.response.reason if self.response.reason else "",
                delay=delay,
            )
            logger.very_verbose_debug(
                "response headers: {headers}", headers=self.response.headers
            )
            data = self.response.read()
            connection.close()
        except socket.timeout:
            if self.silent_fail:
                return None
            logger.warn(
                "connection timeout ({timeout}s)", timeout=self.connection_timeout
            )
            TIMEOUT_Hint.emit()
            return None
        except socket.gaierror as err:
            if self.silent_fail:
                return None
            error_name = "CONNECTION ERROR"
            error_msg = "{err}"
            hint = None
            if str(err) == "[Errno -3] Temporary failure in name resolution":
                padding = "\n" + " " * (len(error_name) + 2)
                error_msg += "{}Could not resolve '{}'".format(padding, domain)
                hint = DNS_Fail_Hint
            elif "[SSL: CERTIFICATE_VERIFY_FAILED]" in str(err):
                hint = SSL_Fail_Hint
            logger.warn(error_msg, err=err, error=error_name)
            if hint is not None:
                hint.emit()
            return None
        except OSError as err:
            if self.silent_fail:
                return None
            logger.warn("{err}", err=err, error="CONNECTION ERROR")
            return None
        return Data(
            data=data,
            code=self.response.status,
            delay=delay,
            reason=self.response.reason,
        )

    def get_last_query_info(self) -> Dict[str, JSONType]:
        base = super().get_last_query_info()
        base.update(self._last_query_info)
        return base


class RedirectFollower(HTTPSLookup[Input, Output]):
    """Follows redirection up to max_depth
    returns final data
    Only works with fixed attributes for domain/path/query..."""

    max_depth = 10
    depth = 0

    def get_data(self) -> Optional[Data]:
        data = super().get_data()
        while data is not None and data.code in [301, 302, 303]:
            if self.response is None:
                return data
            location = self.response.getheader("Location")
            if location is None:
                return data
            self.depth += 1
            logger.debug("Redirect {depth} to : {url}", depth=self.depth, url=location)
            if self.depth >= self.max_depth:
                logger.warn("Redirection depth exceeded")
                return None
            split = normalize_url(location)
            if split is None:
                return data
            self.domain = split[0]
            self.path = split[1]
            data = super().get_data()
        return data


class HTTPSRateCapedLookup(HTTPSLookup[Input, Output]):
    """Add a rate cap to respect polite server requirements"""

    # Time of last query
    # This is a class attribute to the given lookup
    last_query_time: float = 0
    query_delay: float = MIN_QUERY_DELAY  # time between queries, in seconds

    def update_rate_cap(self) -> Optional[float]:
        """Returns the new delay between queries"""
        return None

    def get_data(self) -> Optional[Data]:
        since_last_query = time() - self.last_query_time
        wait = self.query_delay - since_last_query
        if wait > 0.0:
            logger.debug("Rate limier: sleeping for {wait}s", wait=wait)
            sleep(wait)
        self.__class__.last_query_time = time()
        data = super().get_data()
        new_cap = self.update_rate_cap()  # update rate cap with response headers
        if new_cap is not None:
            self.__class__.query_delay = new_cap
        return data



from threading import Condition, Thread
from typing import Callable, Dict, List, Optional, Tuple

from ..bibtex.entry import BibtexEntry
from ..lookups.abstract_base import LookupType
from ..utils.constants import EntryType
from ..utils.logger import logger
from ..utils.safe_json import JSONType


class LookupThread(Thread):
    """As we our I/O limited,
    we can use threads to perform queries
    We create one thread per lookup, to keep query rate polite for each domain"""

    lookup: LookupType
    entries: List[EntryType] = []  # Read only
    condition: Condition
    result: List[
        Tuple[
            Optional[BibtexEntry],
            Dict[str, JSONType],
        ]
    ]  # Write

    position: int
    nb_entries: int

    def __init__(
        self,
        lookup: LookupType,
        entries: List[EntryType],
        condition: Condition,
        bar: Callable[[], None],
    ):
        self.entries = entries
        self.lookup = lookup
        self.condition = condition
        self.position = 0
        self.nb_entries = len(entries)
        self.result = []

        self.bar = bar
        super().__init__(name=lookup.name, daemon=True)

    def run(self) -> None:
        """Starts querying for entries"""
        logger.very_verbose_debug("Starting thread {name}", name=self.name)
        self.condition.acquire()
        while self.position < self.nb_entries:
            lookup = self.lookup(BibtexEntry(self.entries[self.position]))
            self.condition.release()

            result = lookup.query()

            self.condition.acquire()
            self.result.append((result, lookup.get_last_query_info()))
            self.position += 1
            self.bar()
            self.condition.notify()
        self.condition.release()
        return None



"""
Wraps around bibtexparser to provider parser/writer primitives
"""

from pathlib import Path
from typing import List, cast

from bibtexparser.bibdatabase import BibDatabase, UndefinedString
from bibtexparser.bparser import BibTexParser
from bibtexparser.bwriter import BibTexWriter
from bibtexparser.customization import convert_to_unicode

from ..utils.constants import EntryType
from ..utils.logger import logger

writer = BibTexWriter()
writer.indent = "\t"
writer.add_trailing_comma = True
writer.order_entries_by = None  # preserve order
writer.display_order = ("title", "author")


def write(database: BibDatabase) -> str:
    """Transform the database to a bibtex string"""
    return cast(str, writer.write(database).strip() + "\n")


def read(bibtex: str, src: str = "") -> BibDatabase:
    """Parses bibtex string into database"""
    parser = BibTexParser(common_strings=True)
    # Keep non standard entries if present
    parser.ignore_nonstandard_types = False
    try:
        database = parser.parse(bibtex)
    except UndefinedString as err:
        src = " '" + src + "'" if src else ""
        logger.critical(
            "Failed to parse bibtex{src}: {FgPurple}undefined string{Reset} '{err}'",
            src=src,
            err=err,
        )
        exit(1)
    for entry in database.entries:
        convert_to_unicode(entry)
    return database


def file_write(filepath: Path, database: BibDatabase) -> bool:
    """Writes database to given file, stdout if None"""
    output = write(database)
    if filepath is None:
        print(output)
        return True
    try:
        with open(filepath, "w", encoding="utf-8") as file:
            file.write(output)
    except (IOError, UnicodeDecodeError) as err:
        logger.error(
            "Failed to write to '{filepath}' : {FgPurple}{err}{Reset}",
            filepath=str(filepath),
            err=err,
        )
        return False
    return True


def file_read(filepath: Path) -> BibDatabase:
    """reads the given file, parses and normalizes it"""
    # Read and parse the file
    try:
        with open(filepath, "r", encoding="utf-8") as file:
            bibtex = file.read()
    except (IOError, UnicodeDecodeError) as err:
        logger.critical(
            "Failed to read '{filepath}': {FgPurple}{err}{Reset}",
            filepath=str(filepath),
            err=err,
        )
        exit(1)
    return read(bibtex, str(filepath))


def get_entries(db: BibDatabase) -> List[EntryType]:
    """Get entries from a bibdatabase"""
    return list(db.entries)



"""
A class for parsing JSON data without checking structure every time
No operations will raise any error, invalid operations will simply return None
"""


from json import JSONDecodeError, JSONDecoder
from typing import Any, Dict, Iterator, List, Optional, Tuple, TypeVar, Union

JSONType = Union[Dict[str, Any], List[Any], int, float, str, bool, None]
S = TypeVar("S", bound=JSONType)


class SafeJSON:
    """class designed to make failess accesses to a JSON-like structure
    (recursive structure of either Dict[str, SafeJSON], List[SafeJSON], int, float, str, bool, None)

    defines get_item to seamlessly access dict entries (if item is str) or list element (if item is int)
    """

    value: JSONType

    def __init__(self, value: JSONType) -> None:
        self.value = value

    def __getitem__(self, key: Union[int, str]) -> "SafeJSON":
        result: JSONType = None
        if isinstance(key, int):
            if key >= 0 and isinstance(self.value, list) and len(self.value) > key:
                result = self.value[key]
        else:
            if isinstance(self.value, dict):
                result = self.value.get(key)
        return SafeJSON(result)

    @staticmethod
    def from_str(json: str) -> "SafeJSON":
        """Parses a json string into SafeJSON, returns SafeJSON(None) if invalid string"""
        try:
            decoded = JSONDecoder().decode(json)
        except JSONDecodeError:
            return SafeJSON(None)  # empty
        return SafeJSON(decoded)

    @staticmethod
    def from_bytes(json: bytes) -> "SafeJSON":
        """Parses a json bytes string into SafeJSON, returns SafeJSON(None) if invalid string"""
        return SafeJSON.from_str(json.decode())

    def dict_contains(self, key: str) -> bool:
        """Returns true if self is a dict and has the given key"""
        if isinstance(self.value, dict):
            return key in self.value
        return False

    def to_str(self) -> Optional[str]:
        """Returns the value if it is an str, None otherwise"""
        if isinstance(self.value, str):
            return self.value
        return None

    def force_str(self) -> Optional[str]:
        """Returns str(value) if valus is str, int, float or bool"""
        if isinstance(self.value, (str, int, bool, float)):
            return str(self.value)
        return None

    def to_int(self) -> Optional[int]:
        """Returns the value if it is an int, None otherwise"""
        if isinstance(self.value, int):
            return self.value
        return None

    def to_float(self) -> Optional[float]:
        """Returns the value if it is a float, None otherwise"""
        if isinstance(self.value, float):
            return self.value
        return None

    def to_bool(self) -> Optional[bool]:
        """Returns the value if it is a bool, None otherwise"""
        if isinstance(self.value, bool):
            return self.value
        return None

    def to_list(self) -> Optional[List[Any]]:
        """Returns the value if it is an list, None otherwise"""
        if isinstance(self.value, list):
            return self.value
        return None

    def to_dict(self) -> Optional[Dict[str, Any]]:
        """Returns the value if it is a dict, None otherwise"""
        if isinstance(self.value, dict):
            return self.value
        return None

    def iter_list(self) -> "Iterator[SafeJSON]":
        """Iterate through self if it is a list
        else yields nothing"""
        if isinstance(self.value, list):
            for x in self.value:
                yield SafeJSON(x)

    def iter_dict(self) -> "Iterator[Tuple[str, SafeJSON]]":
        """Iterate through self if it is a list
        else yields nothing"""
        if isinstance(self.value, dict):
            for key, val in self.value.items():
                yield key, SafeJSON(val)



"""
A class designed to represent a container defined by either
a list of contained elements or a list of excluded elements
"""


from typing import Callable, Container, Iterable, List, Optional, Set, Tuple, TypeVar

U = TypeVar("U", covariant=True)


T = TypeVar("T")
Q = TypeVar("Q")


class OnlyExclude(Container[T]):
    """Class to represent a set defined by either
    - only containing elements
    - not containing elements
    Implement the in operator
    and a filter iterator"""

    onlys: Optional[List[T]]
    nots: Optional[List[T]]

    def __init__(self, onlys: Optional[List[T]], nots: Optional[List[T]]) -> None:
        """Create a new instance with onlys or nots.
        If both are specified, onlys takes precedence.
        Not that an empty container is not None,
        so an empty onlys will create a container containing nothing"""

        self.onlys = onlys
        self.nots = nots

    @classmethod
    def from_nonempty(cls, onlys: List[T], nots: List[T]) -> "OnlyExclude[T]":
        """A different initializer, which considers empty containers to be None"""
        o = onlys if len(onlys) > 0 else None
        n = nots if len(nots) > 0 else None
        return cls(o, n)  # We need the class for type instanciation

    def __contains__(self, obj: T) -> bool:  # type: ignore[override]
        """Check if obj is valid given the exclusion rules"""
        if self.onlys is not None:
            return obj in self.onlys
        if self.nots is not None:
            return obj not in self.nots
        return True

    def filter(self, iterable: Iterable[Q], map: Callable[[Q], T]) -> Iterable[Q]:
        """Returns a filtered Iterator
        Note that this filter is consumed after the first use"""
        return filter(lambda x: map(x) in self, iterable)

    def unused(self, iterable: Iterable[T]) -> Tuple[Set[T], Set[T]]:
        """Return set of unused filters:
        - set of unused only filters
        - set of unused not filters"""
        if self.onlys is not None:
            unused = set(self.onlys)
            for x in iterable:
                if x in self.onlys:
                    unused.discard(x)
            nots = set() if self.nots is None else set(self.nots)
            return unused, nots
        if self.nots is not None:
            unused = set(self.nots)
            for x in iterable:
                if x in self.nots:
                    unused.discard(x)
            return set(), unused
        return set(), set()



from pathlib import Path
from sys import stdout
from typing import List, Optional

from ..APIs import LOOKUP_NAMES, LOOKUPS
from ..bibtex.io import writer
from ..lookups.condition_mixin import FieldConditionMixin
from ..lookups.https import HTTPSLookup
from ..utils.ansi import ANSICodes, ansi_format
from ..utils.constants import (
    CONNECTION_TIMEOUT,
    LICENSE,
    SCRIPT_NAME,
    URL,
    VERSION_DATE,
    VERSION_STR,
)
from ..utils.logger import logger
from ..utils.only_exclude import OnlyExclude
from .autocomplete import BibtexAutocomplete
from .parser import (
    HELP_TEXT,
    flatten,
    get_bibfiles,
    indent_string,
    make_output_names,
    parser,
)


def main(argv: Optional[List[str]] = None) -> None:
    """The main function of bibtexautocomplete
    Takes an argv like List as argument,
    if none, uses sys.argv
    see HELP_TEXT or main(["-h"]) for details"""
    if argv is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(argv)

    ANSICodes.use_ansi = stdout.isatty() and not args.no_color

    if args.help:
        print(
            ansi_format(
                HELP_TEXT,
                TIMEOUT=CONNECTION_TIMEOUT,
                VERSION=VERSION_STR,
                VERSION_DATE=VERSION_DATE,
                LOOKUPS=LOOKUP_NAMES,
                NAME=SCRIPT_NAME,
                URL=URL,
                LICENSE=LICENSE,
            )
        )
        return
    if args.version:
        print(
            "{NAME} version {VERSION} ({VERSION_DATE})".format(
                NAME=SCRIPT_NAME, VERSION=VERSION_STR, VERSION_DATE=VERSION_DATE
            )
        )
        return

    if args.silent:
        args.verbose = -args.silent
    logger.set_verbosity(args.verbose)

    args.input = flatten(args.input)
    # No input -> CWD
    if args.input == []:
        args.input = [Path(".")]
    args.input = flatten(map(get_bibfiles, args.input))
    if args.inplace:
        if args.output != []:
            logger.warn("Inplace mode, ignoring specified output files")
        args.output = args.input
    else:
        args.output = make_output_names(args.input, args.output)

    writer.align_values = args.align_values
    writer.comma_first = args.comma_first
    writer.add_trailing_comma = args.no_trailing_comma
    writer.indent = indent_string(args.indent)

    HTTPSLookup.connection_timeout = args.timeout if args.timeout > 0.0 else None
    HTTPSLookup.ignore_ssl = args.ignore_ssl
    lookups = (
        OnlyExclude[str]
        .from_nonempty(args.only_query, args.dont_query)
        .filter(LOOKUPS, lambda x: x.name)
    )
    fields = OnlyExclude[str].from_nonempty(args.only_complete, args.dont_complete)
    entries = OnlyExclude[str].from_nonempty(args.only_entry, args.exclude_entry)

    FieldConditionMixin.fields_to_complete = set(
        fields.filter(FieldConditionMixin.fields_to_complete, lambda x: x)
    )

    databases = BibtexAutocomplete.read(args.input)
    completer = BibtexAutocomplete(
        databases, lookups, fields, entries, args.force_overwrite, args.prefix
    )
    completer.print_filters()
    try:
        completer.autocomplete(args.verbose < 0)
        completer.print_changes()
        if args.dump_data is not None:
            completer.write_dumps(args.dump_data)
        if not args.no_output:
            completer.write(args.output)
    except KeyboardInterrupt:
        logger.warn("Interrupted")



"""
Lookup info from https://arxiv.org
"""

from re import sub
from typing import Dict, Iterable, List, Optional, Tuple
from xml.etree.ElementTree import Element, ParseError, fromstring

from ..bibtex.author import Author
from ..bibtex.entry import BibtexEntry, FieldNames
from ..lookups.lookups import XML_T_Lookup
from ..utils.constants import QUERY_MAX_RESULTS


class ArxivLookup(XML_T_Lookup):
    """Lookup info on https://arxiv.org
    Uses the arXiv REST API, documented here:
    https://arxiv.org/help/api/user-manual

    example URLs:
    https://export.arxiv.org/api/query?search_query=Reactive+Path+Deformation+for+Nonholonomic+Mobile+Robots&start=0&max_results=3
    """

    name = "arxiv"

    domain = "export.arxiv.org"
    path = "/api/query"
    accept = "application/xml"

    safe = ":"

    query_delay = 3.0  # seconds

    xml_prefix = "{http://www.w3.org/2005/Atom}"

    def get_params(self) -> Dict[str, str]:
        if self.title is None:
            raise ValueError("arXiv called with no title")
        return {
            "search_query": f'ti:"{self.title}"',
            "start": "0",
            "max_results": str(QUERY_MAX_RESULTS),
        }

    def get_results(self, data: bytes) -> Optional[Iterable[Element]]:
        """Return the result list"""
        try:
            xml = fromstring(data.decode())
        except ParseError:
            return None
        return xml.findall(self.xml_prefix + "entry")

    def xml_gettext(self, elem: Element, attr: str) -> Optional[str]:
        value = elem.find(self.xml_prefix + attr)
        if value is None:
            return None
        return value.text

    def get_title(self, elem: Element) -> Optional[str]:
        title = self.xml_gettext(elem, "title")
        if title is not None:
            return sub(r"\s+", " ", title)
        return None

    def get_doi(self, elem: Element) -> Optional[str]:
        """Tries to find DOI"""
        for doi in elem.findall("{http://arxiv.org/schemas/atom}doi"):
            if doi.text:
                return doi.text
        for link in elem.findall("{http://www.w3.org/2005/Atom}link"):
            if link.attrib.get("title") == "doi" and link.text:
                return link.text
        return None

    def get_authors(self, elem: Element) -> List[Author]:
        authors = elem.findall(self.xml_prefix + "author")
        formatted = []
        for author in authors:
            name = Author.from_name(self.xml_gettext(author, "name"))
            if name is not None:
                formatted.append(name)
        return formatted

    def get_date(self, elem: Element) -> Tuple[Optional[str], Optional[str]]:
        date = self.xml_gettext(elem, "published")
        year: Optional[str] = None
        month: Optional[str] = None
        if isinstance(date, str):
            if len(date) >= 4 and date[0:4].isnumeric():
                year = date[0:4]
                if len(date) >= 7 and date[5:7].isnumeric():
                    month = date[5:7]
        return year, month

    def get_link(self, elem: Element) -> Optional[str]:
        for link in elem.findall("{http://www.w3.org/2005/Atom}link"):
            if link.attrib.get("title") == "pdf" and link.text:
                return link.text
        return None

    def get_value(self, result: Element) -> BibtexEntry:
        """Extract bibtex data from JSON output"""
        year, month = self.get_date(result)
        values = BibtexEntry()
        values.author = self.get_authors(result)
        values.doi = self.get_doi(result)
        values.month = month
        values.title = self.get_title(result)
        values.url = self.get_link(result)
        values.year = year
        return values

    fields = {
        FieldNames.AUTHOR,
        FieldNames.DOI,
        FieldNames.MONTH,
        FieldNames.TITLE,
        FieldNames.URL,
        FieldNames.YEAR,
    }



"""
Lookup info from https://researchr.org/
"""

from typing import Iterable, List
from urllib.parse import quote_plus

from ..bibtex.author import Author
from ..bibtex.entry import BibtexEntry, FieldNames
from ..lookups.lookups import JSON_AT_Lookup
from ..utils.safe_json import SafeJSON


class ResearchrLookup(JSON_AT_Lookup):
    """Lookup for info on https://researchr.org/
    Uses the API documented here:
    https://researchr.org/about/api

    Example urls:
    No doi mode.
    Author title mode:
    https://researchr.org/api/search/publication/Lamiraux+Reactive+Path+Deformation+for+Nonholonomic+Mobile+Robots
    """

    name = "researchr"

    domain = "researchr.org"
    path = "/api/search/publication/"

    def get_path(self) -> str:
        search = ""
        if self.author is not None:
            search += self.author + " "
        if self.title is not None:
            search += self.title + " "
        return self.path + quote_plus(search.strip())

    def get_results(self, data: bytes) -> Iterable[SafeJSON]:
        """Return the result list"""
        return SafeJSON.from_bytes(data)["result"].iter_list()

    @staticmethod
    def get_authors(authors: SafeJSON) -> List[Author]:
        """Return a bibtex formatted list of authors"""
        formatted = []
        for author in authors.iter_list():
            aut = Author.from_name(author["alias"]["name"].to_str())
            if aut is not None:
                formatted.append(aut)
        return formatted

    def get_value(self, result: SafeJSON) -> BibtexEntry:
        page_1 = result["firstpage"].to_str()
        page_n = result["lastpage"].to_str()
        values = BibtexEntry()
        values.address = result["address"].to_str()
        values.author = self.get_authors(result["authors"])
        values.booktitle = result["booktitle"].to_str()
        values.doi = result["doi"].to_str()
        values.editor = self.get_authors(result["editors"])
        values.month = result["month"].to_str()
        values.number = result["number"].to_str()
        values.organization = result["organization"].to_str()
        values.pages = (
            f"{page_1}-{page_n}" if page_1 is not None and page_n is not None else None
        )
        values.publisher = result["publisher"].to_str()
        values.title = result["title"].to_str()
        values.volume = result["volume"].to_str()
        values.year = result["year"].to_str()
        return values

    fields = {
        FieldNames.ADDRESS,
        FieldNames.AUTHOR,
        FieldNames.BOOKTITLE,
        FieldNames.DOI,
        FieldNames.EDITOR,
        FieldNames.MONTH,
        FieldNames.NUMBER,
        FieldNames.ORGANIZATION,
        FieldNames.PAGES,
        FieldNames.PUBLISHER,
        FieldNames.TITLE,
        FieldNames.VOLUME,
        FieldNames.YEAR,
    }



"""
Queries to https://doi.org/<doi>
used to check that dois are valid
"""

from typing import Optional
from urllib.parse import quote, urlencode

from ..bibtex.normalize import normalize_doi, normalize_str_weak, normalize_url
from ..lookups.abstract_base import Data
from ..lookups.condition_mixin import ConditionMixin
from ..lookups.https import HTTPSRateCapedLookup, RedirectFollower
from ..utils.safe_json import SafeJSON


class URLCheck(
    ConditionMixin[str, Optional[Data]], RedirectFollower[str, Optional[Data]]
):
    """Checks that an URL exists (should return 200)
    Follows redirection (up to a certain depth)"""

    name = "url_checker"
    accept = "text/html"
    silent_fail = True

    def condition(self) -> bool:
        return self.is_valid

    def __init__(self, input: str):
        split = normalize_url(input)
        if split is None:
            self.is_valid = False
        else:
            self.is_valid = True
            self.domain = split[0]
            self.path = split[1]

    def process_data(self, data: Data) -> Optional[Data]:
        if data.code != 200:
            return None
        return data


class DOICheck(
    ConditionMixin[str, Optional[bool]],
    HTTPSRateCapedLookup[str, Optional[bool]],
):
    """
    Queries https://doi.org/api/handles/<doi> to check DOI exists
    and get redirection URL, then checks the redirection URL returns 200
    or redirects to an URL that does.

    Example URL: https://doi.org/api/handles/10.1109/tro.2004.829459
    """

    name = "doi_checker"
    silent_fail = True

    params = {"type": "URL"}
    domain = "doi.org"
    path = "/api/handles/"

    not_available_checks = [
        "not available",
        "not found",
    ]

    doi: str

    def __init__(self, input: str) -> None:
        self.input = input

    def condition(self) -> bool:
        """Checks that a doi is valid"""
        doi = normalize_doi(self.input)
        if doi is not None:
            self.doi = doi
            return True
        return False

    def get_path(self) -> str:
        return self.path + quote(self.doi) + "?" + urlencode(self.params)

    def process_data(self, data: Data) -> Optional[bool]:
        if data.code != 200:
            return None
        json = SafeJSON.from_bytes(data.data)
        if json["responseCode"].to_int() != 1:
            return None
        for value in json["values"].iter_list():
            if value["type"].to_str() == "URL":
                if self.check_url(value["data"]["value"].to_str()):
                    return True
        return None

    def check_url(self, url: Optional[str]) -> bool:
        if url is not None:
            checker = URLCheck(url)
            final = checker.query()
            if final is not None:
                text = normalize_str_weak(final.data.decode())
                # Some website, namely springer, don't send 404 for invalid DOIs...
                # See: https://link.springer.com/deleted
                for elem in self.not_available_checks:
                    if elem in text:
                        return False
                return True
        return False



"""
Function to determine how likely two entries are the same
"""

from typing import Set, Tuple

from .entry import BibtexEntry
from .normalize import normalize_str, normalize_str_weak

ENTRY_CERTAIN_MATCH = 1000  # Maximum score
ENTRY_NO_MATCH = 0  # Minimum score


def author_set(entry: BibtexEntry) -> Set[str]:
    """Returns the set of normalized author lastname from entry"""
    authors = entry.author
    last_names: Set[str] = set()
    for author in authors:
        normalized = normalize_str(author.lastname).replace(" ", "")
        if normalized != "":
            last_names.add(normalized)
    return last_names


def common_authors(a: BibtexEntry, b: BibtexEntry) -> Tuple[int, int, int]:
    """Returns:
    number of common authors, number of a authors, number of authors of b"""
    authors_a = author_set(a)
    authors_b = author_set(b)
    common = authors_a.intersection(authors_b)
    return len(common), len(authors_a), len(authors_b)


def match_score(a: BibtexEntry, b: BibtexEntry) -> int:
    """
    Assign a score between ENTRY_NO_MATCH and ENTRY_CERTAIN_MATCH (included)
    representing how likely two entries are to be matched
    """
    if a.doi is not None and a.doi == b.doi:
        # DOIs are unique => entries match
        return ENTRY_CERTAIN_MATCH
    if a.title is not None and b.title is not None:
        # Title is our best identifier
        if normalize_str_weak(a.title) == normalize_str_weak(b.title):
            score = ENTRY_CERTAIN_MATCH // 2
        elif normalize_str(a.title) == normalize_str(b.title):
            score = ENTRY_CERTAIN_MATCH // 3
        else:
            return ENTRY_NO_MATCH
        authors_common, authors_a, authors_b = common_authors(a, b)
        if authors_common == 0 and authors_a > 0 and authors_b > 0:
            # No common authors despite some authors being known on both sides
            return ENTRY_NO_MATCH
        if authors_common != 0:
            if authors_common == authors_a and authors_common == authors_b:
                score += ENTRY_CERTAIN_MATCH // 3
            elif authors_common == authors_a or authors_common == authors_b:
                score += ENTRY_CERTAIN_MATCH // 6
            else:
                score += ENTRY_CERTAIN_MATCH // 9
        # Could check additional fields here
        return score
    return ENTRY_NO_MATCH



"""
Combine all the mixins into full fledge classes
"""

from xml.etree.ElementTree import Element

from ..bibtex.entry import BibtexEntry
from ..utils.safe_json import SafeJSON
from .condition_mixin import FieldConditionMixin
from .https import HTTPSRateCapedLookup
from .multiple_mixin import (
    DATQueryMixin,
    DTQueryMixin,
    TitleAuthorQueryMixin,
    TitleQueryMixin,
)
from .search_mixin import EntryMatchSearchMixin


class JSON_DAT_Lookup(
    FieldConditionMixin,
    DATQueryMixin,
    EntryMatchSearchMixin[SafeJSON],
    HTTPSRateCapedLookup[BibtexEntry, BibtexEntry],
):
    pass


class JSON_DT_Lookup(
    FieldConditionMixin,
    DTQueryMixin,
    EntryMatchSearchMixin[SafeJSON],
    HTTPSRateCapedLookup[BibtexEntry, BibtexEntry],
):
    pass


class JSON_AT_Lookup(
    FieldConditionMixin,
    TitleAuthorQueryMixin,
    EntryMatchSearchMixin[SafeJSON],
    HTTPSRateCapedLookup[BibtexEntry, BibtexEntry],
):
    pass


class XML_T_Lookup(
    FieldConditionMixin,
    TitleQueryMixin,
    EntryMatchSearchMixin[Element],
    HTTPSRateCapedLookup[BibtexEntry, BibtexEntry],
):
    pass



"""
Mixin that override process_data
assuming there are multiple results

New virtual methods:
- get_results : bytes -> Optional[Iterable[result]]
    process data into a list of results
- get_value : result -> BibtexEntry - builds value from a result
"""

from typing import Generic, Iterable, List, Optional, TypeVar

from ..bibtex.entry import BibtexEntry
from ..bibtex.matching import ENTRY_CERTAIN_MATCH, ENTRY_NO_MATCH, match_score
from ..utils.logger import logger
from .abstract_base import AbstractEntryLookup, Data

result = TypeVar("result")


class SearchResultMixin(Generic[result]):
    """Iterates through multiple results until a matching one is found

    Defines:
    - process_data : Data -> Optional[BibtexEntry]

    Virtual methods defined here:
    - get_results : bytes -> Optional[Iterable[result]]
        process data into a list of results
    - get_value : result -> Optional[BibtexEntry] - builds value from a result
    - match_score : entry -> result -> int - matching score
        between the entry and our search term
        value between ENTRY_NO_MATCH and ENTRY_CERTAIN_MATCH (included)

    Attribute: ok_codes : list of expected codes, fails data.code is not in them
      defaults to [200]
    """

    ok_codes: List[int] = [200]

    def get_results(self, data: bytes) -> Optional[Iterable[result]]:
        """Parse the data into a list of results to check
        Return None if no results/invalid data"""
        raise NotImplementedError("should be overridden in child class")

    def get_value(self, res: result) -> BibtexEntry:
        """Return the relevant value (e.g. updated entry)"""
        raise NotImplementedError("should be overridden in child class")

    def match_score(self, entry: BibtexEntry, res: result) -> int:
        """
        Assign a score between ENTRY_NO_MATCH and ENTRY_CERTAIN_MATCH (included)
        representing how likely the given entries matches our search term
        - entry:BibtexEntry is self.get_value(res)
        - res:result is also passed in case get_value forgets some data.
        """
        raise NotImplementedError("should be overridden in child class")
        return match_score(self.entry, entry)

    def process_data(self, data: Data) -> Optional[BibtexEntry]:
        """Iterate through results until one matches"""
        if data.code not in self.ok_codes:
            logger.warn(
                "response: {FgYellow}{status}{reason}{Reset} in {delay}s",
                status=data.code,
                reason=" " + data.reason if data.reason else "",
                delay=data.delay,
            )
            return None
        results = self.get_results(data.data)
        if results is None:
            logger.verbose_debug("no results")
            return None
        max_score = ENTRY_NO_MATCH
        max_entry: Optional[BibtexEntry] = None
        for res in results:
            entry = self.get_value(res)
            score = self.match_score(entry, res)
            logger.verbose_debug("match {} for {}", score, entry._entry)
            if score >= ENTRY_CERTAIN_MATCH:
                return entry
            if score > max_score:
                max_score = score
                max_entry = entry
        return max_entry


class EntryMatchSearchMixin(SearchResultMixin[result], AbstractEntryLookup):
    """Uses the bibtex entry matcher to get match score"""

    def match_score(self, entry: BibtexEntry, _res: result) -> int:
        return match_score(self.entry, entry)



"""
A class to represent author/editor names and read/write them to valid bibtex
"""

from typing import Any, List, Optional

AUTHOR_JOIN = " and "


class Author:
    firstnames: Optional[str]
    lastname: str

    def __init__(self, lastname: str, firstnames: Optional[str]) -> None:
        self.lastname = lastname
        self.firstnames = firstnames

    def __repr__(self) -> str:
        return f"Author({self.lastname}, {self.firstnames})"

    def to_bibtex(self) -> str:
        """Returns a bibtex representation of self:
        lastname, firstname"""
        if self.firstnames is not None:
            return f"{self.lastname}, {self.firstnames}"
        return self.lastname

    @staticmethod
    def list_to_bibtex(authors: "List[Author]") -> str:
        return AUTHOR_JOIN.join(author.to_bibtex() for author in authors)

    def __eq__(self, other: Any) -> bool:
        "Used in test only"
        if not isinstance(other, Author):
            return False
        return self.firstnames == other.firstnames and self.lastname == other.lastname

    @staticmethod
    def from_name(name: Optional[str]) -> "Optional[Author]":
        """Reads a bibtex string into a author name"""
        if name is None or name == "" or name.isspace():
            return None
        name = name.replace("\n", "").strip()
        if "," in name:
            namesplit = name.split(",", 1)
            last = namesplit[0].strip()
            firsts = [i.strip() for i in namesplit[1].split()]
        else:
            namesplit = name.split()
            last = namesplit.pop()
            firsts = [i.replace(".", ". ").strip() for i in namesplit]
        if last in ["jnr", "jr", "junior"]:
            last = firsts.pop()
        for item in firsts:
            if item in ["ben", "van", "der", "de", "la", "le"]:
                last = firsts.pop() + " " + last
        first = " ".join(firsts) if firsts else None
        return Author(last, first)

    @classmethod
    def from_namelist(cls, authors: str) -> "List[Author]":
        """Return a list of 'first name', 'last name' for authors"""
        result = []
        for name in authors.replace("\n", " ").replace("\t", " ").split(AUTHOR_JOIN):
            aut = cls.from_name(name)
            if aut is not None:
                result.append(aut)
        return result



"""
Functions used to normalize bibtex fields
"""

import unicodedata
from datetime import date
from re import compile, search, sub
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlsplit

from ..utils.constants import EntryType


def make_plain(value: Optional[str]) -> Optional[str]:
    """Returns a plain version of the field (remove redundant braces)
    returns None if the field is None or empty string"""
    if value is not None:
        plain = value.replace("{", "").replace("}", "").strip()
        if plain != "" and not plain.isspace():
            return plain
    return None


def has_data(value: Optional[str]) -> bool:
    """Return true if the given value contains data, false otherwise"""
    return make_plain(value) is not None


def get_field(entry: EntryType, field: str) -> Optional[str]:
    """Check if given field exists and is non-empty
    if so, removes braces and returns it"""
    if field in entry:
        return make_plain(entry[field])
    return None


def has_field(entry: EntryType, field: str) -> bool:
    """Check if a given entry has non empty field"""
    return has_data(get_field(entry, field))


def strip_accents(string: str) -> str:
    """replace accented characters with their non-accented variants"""
    # Solution from https://stackoverflow.com/a/518232
    return "".join(
        c
        for c in unicodedata.normalize("NFD", string)
        if unicodedata.category(c) != "Mn"
    )


def normalize_str_weak(string: str) -> str:
    """Converts to lower case, strips accents,
    replace tabs and newline with spaces,
    removes duplicate spaces"""
    string = strip_accents(string).lower()
    return sub(r"\s+", " ", string)


def normalize_str(string: str) -> str:
    """Normalize string for decent comparison
    Converts to lower case, strips accents
    Replaces all non alpha-numeric characters with spaces
    Removes duplicate spaces"""
    res = ""
    prev_space = False
    for x in strip_accents(string):
        if x.isalnum():
            res += x
            prev_space = False
        elif not prev_space:
            res += " "
            prev_space = True
    return res.lower().strip()


# 100 most common english words, from Wikipedia
COMMON_WORDS = [
    "the",
    "be",
    "to",
    "of",
    "and",
    "a",
    "in",
    "that",
    "have",
    "i",
    "it",
    "for",
    "not",
    "on",
    "with",
    "he",
    "as",
    "you",
    "do",
    "at",
    "this",
    "but",
    "his",
    "by",
    "from",
    "they",
    "we",
    "say",
    "her",
    "she",
    "or",
    "an",
    "will",
    "my",
    "one",
    "all",
    "would",
    "there",
    "their",
    "what",
    "so",
    "up",
    "out",
    "if",
    "about",
    "who",
    "get",
    "which",
    "go",
    "me",
    "when",
    "make",
    "can",
    "like",
    "time",
    "no",
    "just",
    "him",
    "know",
    "take",
    "people",
    "into",
    "year",
    "your",
    "good",
    "some",
    "could",
    "them",
    "see",
    "other",
    "than",
    "then",
    "now",
    "look",
    "only",
    "come",
    "its",
    "over",
    "think",
    "also",
    "back",
    "after",
    "use",
    "two",
    "how",
    "our",
    "work",
    "first",
    "well",
    "way",
    "even",
    "new",
    "want",
    "because",
    "any",
    "these",
    "give",
    "day",
    "most",
    "us",
]


def keywords(title: str) -> List[str]:
    """Returns only the best keywords in the given title"""
    title = normalize_str(title)
    words = []
    for word in title.split(" "):
        if word not in COMMON_WORDS:
            words.append(word)
    return words


DOI_REGEX = r"(10\.\d{4,5}\/[\S]+[^;,.\s])$"


def normalize_doi(doi_or_url: Optional[str]) -> Optional[str]:
    """Returns doi to canonical form (i.e. removing url)"""
    if doi_or_url is not None:
        match = search(DOI_REGEX, doi_or_url)
        if match is not None:
            return match.group(1).lower()  # DOI's are case insensitive
    return None


def months_format(month: int, format: str) -> str:
    """Localized month format"""
    return date(2001, month, 1).strftime(format)


def get_locale_months() -> Dict[str, int]:
    mapping = {}
    for month in range(1, 13):
        for format in ("%B", "%b"):  # "%m", "%-m"
            mapping[months_format(month, format).lower()] = month
    return mapping


EN_MONTHS = {
    "january": 1,
    "jan": 1,
    "01": 1,
    "1": 1,
    "february": 2,
    "feb": 2,
    "02": 2,
    "2": 2,
    "march": 3,
    "mar": 3,
    "03": 3,
    "3": 3,
    "april": 4,
    "apr": 4,
    "04": 4,
    "4": 4,
    "may": 5,
    "05": 5,
    "5": 5,
    "june": 6,
    "jun": 6,
    "06": 6,
    "6": 6,
    "july": 7,
    "jul": 7,
    "07": 7,
    "7": 7,
    "august": 8,
    "aug": 8,
    "08": 8,
    "8": 8,
    "september": 9,
    "sep": 9,
    "09": 9,
    "9": 9,
    "october": 10,
    "oct": 10,
    "10": 10,
    "november": 11,
    "nov": 11,
    "11": 11,
    "december": 12,
    "dec": 12,
    "12": 12,
}


def normalize_month(month: str) -> str:
    """Tries to normalize a month to it's number "1" to "12"
    returns month unchanged if unsuccessful"""
    months = EN_MONTHS.copy()
    months.update(get_locale_months())
    norm = normalize_str(month)
    if norm in months:
        return str(months[norm])
    return month


URL_REGEX = compile(
    "((http|https)://)(www.)?"
    + "[a-zA-Z0-9@:%._\\+~#?&//=]"
    + "{2,256}\\.[a-z]"
    + "{2,6}\\b([-a-zA-Z0-9@:%"
    + "._\\+~#?&//=]*)"
)


def normalize_url(url: str) -> Optional[Tuple[str, str]]:
    """Splits and url into domain/path
    Returns none if url is not valid"""
    if not search(URL_REGEX, url):
        return None
    split = urlsplit(url)
    if split.netloc == "":
        return None
    domain = split.netloc
    path = split.path
    if split.query != "":
        path += "?" + split.query
    return domain, path



"""
Mixin to check if a condition holds before performing queries
"""

from typing import Optional, Set

from ..bibtex.entry import BibtexEntry, SearchedFields
from .abstract_base import AbstractEntryLookup, AbstractLookup, Input, Output


class ConditionMixin(AbstractLookup[Input, Output]):
    """Mixin to query only if a condition holds,

    inherit from this before the base Lookup class
    e.g. class MyLookup(..., ConditionMixin, ..., MyLookup):

    Adds the condition : Self -> bool method (default always True)"""

    def condition(self) -> bool:
        """override this to check a condition before
        performing any queries"""
        return True

    def query(self) -> Optional[Output]:
        """calls parent query only if condition is met"""
        if self.condition():
            return super().query()
        return None


class FieldConditionMixin(
    ConditionMixin[BibtexEntry, BibtexEntry], AbstractEntryLookup
):
    """Mixin used to query only if there exists a field in self.fields
    that does not exists in self.entry

    inherit from this before the base class
    e.g. class MyLookup(..., FieldConditionMixin, ..., MyLookup):

    Virtual attribute:
    - fields : Iterable[str] - list of fields that can be added to an entry by this lookup
    """

    # list of fields that can be added to an entry by this lookup
    fields: Set[str]

    fields_to_complete: Set[str] = SearchedFields

    def condition(self) -> bool:
        """Only return True if there exists a field in self.fields
        that is not in self.entry"""
        for field in self.fields.intersection(self.fields_to_complete):
            if field not in self.entry:
                return True
        return False



"""
Command-line argument parser
"""

from argparse import ArgumentParser
from os import listdir
from pathlib import Path
from typing import Iterable, List, TypeVar

from ..APIs import LOOKUP_NAMES
from ..bibtex.entry import FieldNamesSet
from ..utils.constants import BTAC_FILENAME, CONNECTION_TIMEOUT, SCRIPT_NAME
from ..utils.logger import logger

T = TypeVar("T")


def flatten(list_of_lists: Iterable[List[T]]) -> List[T]:
    """flatten a nested list"""
    return [val for sublist in list_of_lists for val in sublist]


def make_output_name(input: Path) -> Path:
    """Returns the new renamed path
    e.g. example.bib -> example.btac.bib"""
    name = input.name
    suffix = ""
    if "." in name:
        split = name.split(".")
        suffix = "." + split.pop()
        name = ".".join(split)
    return Path(input.parent, BTAC_FILENAME.format(name=name, suffix=suffix))


def make_output_names(inputs: List[Path], outputs: List[Path]) -> List[Path]:
    """Returns output names
    - the first ones are taken from outputs
    - if outputs < inputs, uses inputs with renaming xxx.bib -> xxx.btac.bib
    - if inputs < outputs, issues a warning"""
    len_inputs = len(inputs)
    len_outputs = len(outputs)
    if len_outputs > len_inputs:
        logger.warn(
            "Too many output files specified: got {outs} for {ins} input files",
            outs=len_outputs,
            ins=len_inputs,
        )
    for ii in range(len_outputs, len_inputs, 1):
        outputs.append(make_output_name(inputs[ii]))
    return outputs


def indent_string(indent: str) -> str:
    """Reads and formats an ident string
    whitespace -> whitespace
    "\\t" -> tab
    number -> number of spaces"""
    if indent.isnumeric():
        return " " * int(indent)
    sane = indent.replace("t", "\t").replace("n", "\n").replace("_", " ")
    if not (sane.isspace() or sane == ""):
        logger.critical(
            (
                "--fi/--indent should be a number or string "
                "with spaces, '_', 't' and 'n' only.\nGot: '{}'"
            ).format(indent)
        )
        exit(5)
    return sane


def filter_bibs(files: List[Path]) -> List[Path]:
    """Filter for files ending in .bib
    Ignores generated ".btac.bib" files unless they are the only ones present"""
    bibs = []
    btac_bibs = []
    for file in files:
        if str(file).endswith(BTAC_FILENAME.format(name="", suffix=".bib")):
            btac_bibs.append(file)
        elif file.suffix == ".bib":
            bibs.append(file)
    if bibs:
        return bibs
    return btac_bibs


def get_bibfiles(input: Path) -> List[Path]:
    """Finds bibfiles contained in a folder if given as inputt"""
    if not input.is_dir():
        return [input]
    try:
        files = [input / x for x in listdir(input) if (input / x).is_file()]
    except IOError as err:
        logger.critical(
            "Failed to read '{filepath}': {FgPurple}{err}{Reset}",
            filepath=str(input),
            err=err,
        )
        exit(1)
    return filter_bibs(files)


parser = ArgumentParser(
    prog=SCRIPT_NAME,
    add_help=False,
    usage="btac [--options] <input_files>\nSee help for a list of options.\n",
)

FIELD_NAMES = sorted(FieldNamesSet)

parser.add_argument(
    "--dont-query", "-Q", action="append", default=[], choices=LOOKUP_NAMES
)
parser.add_argument(
    "--only-query", "-q", action="append", default=[], choices=LOOKUP_NAMES
)
parser.add_argument(
    "--dont-complete", "-C", action="append", default=[], choices=FIELD_NAMES
)
parser.add_argument(
    "--only-complete",
    "-c",
    action="append",
    default=[],
    choices=FIELD_NAMES,
)
parser.add_argument("--exclude-entry", "-E", action="append", default=[])
parser.add_argument("--only-entry", "-e", action="append", default=[])

parser.add_argument("--align-values", "--fa", action="store_true")
parser.add_argument("--comma-first", "--fc", action="store_true")
parser.add_argument("--no-trailing-comma", "--fl", action="store_false")
parser.add_argument("--indent", "--fi", default="\t")

parser.add_argument("--force-overwrite", "-f", action="store_true")
parser.add_argument("--prefix", "-p", action="store_true")
parser.add_argument("--inplace", "-i", action="store_true")
parser.add_argument("--timeout", "-t", type=float, default=CONNECTION_TIMEOUT)
parser.add_argument("--verbose", "-v", action="count", default=0)
parser.add_argument("--silent", "-s", action="store_true")
parser.add_argument("--no-color", "-n", action="store_true")
parser.add_argument("--ignore-ssl", "-S", action="store_true")

parser.add_argument("--version", action="store_true")
parser.add_argument("--help", "-h", action="store_true")

parser.add_argument("--dump-data", "-d", type=Path)
parser.add_argument("--no-output", "-O", action="store_true")
parser.add_argument("--output", "-o", type=Path, action="append", default=[])
parser.add_argument("input", nargs="*", type=Path, action="append", default=[])

HELP_TEXT = """{StBold}{FgYellow}{NAME}{Reset} {StBold}version {VERSION} ({VERSION_DATE}){Reset}

BibTex AutoComplete is a command line script to find and add missing fields (doi, publisher...) to
a bibtex bibliography by searching online databases. It gets data by searching:
  {LOOKUPS}

More information and demo:
  {StUnderline}{URL}{Reset}

{StBold}Usage:{Reset}
  {StBold}{FgYellow}{NAME}{Reset} {FgYellow}[--flags] <input_files>{Reset}

{StBold}Examples:{Reset}
  {StBold}{FgYellow}{NAME}{Reset} {FgYellow}my_bib.bib{Reset}         writes to my_bib.btac.bib
  {StBold}{FgYellow}{NAME}{Reset} {FgYellow}-i my_bib.bib{Reset}      inplace modify
  {StBold}{FgYellow}{NAME}{Reset} {FgYellow}a.bib -o b.bib c.bib -o d.bib e.bib{Reset}
      writes completed a.bib in b.bib; c.bib in d.bib; and e.bib in e.btac.bib
  {StBold}{FgYellow}{NAME}{Reset} {FgYellow}dir/{Reset}            completes all '.bib' files in dir,
      ignores any '.btac.bib' files, not recursive on subfolders.
  {StBold}{FgYellow}{NAME}{Reset}                    same as 'btac .'

{StBold}Optional arguments:{Reset} can all be used multiple times
  {FgYellow}-o --output{Reset} {FgGreen}<file.bib>{Reset}      Write output to given file
        With multiple input/outputs they are mapped in appearance order
        Extra inputs use default output name <filename>.btac.bib

  {FgYellow}-q --only-query{Reset} {FgGreen}<website>{Reset}   Only query the given sites
  {FgYellow}-Q --dont-query{Reset} {FgGreen}<website>{Reset}   Don't query the given sites
        Website must be one of: {LOOKUPS}

  {FgYellow}-e --only-entry{Reset}    {FgGreen}<id>{Reset}     Only perform lookup these entries
  {FgYellow}-E --exclude-entry{Reset} {FgGreen}<id>{Reset}     Don't perform lookup these entries
        ID is the identifier in bibtex (e.g. @inproceedings{{<id> ... }})

  {FgYellow}-c --only-complete{Reset} {FgGreen}<field>{Reset}  Only complete the given fields
  {FgYellow}-C --dont-complete{Reset} {FgGreen}<field>{Reset}  Don't complete the given fields
        Field is a bibtex field (e.g. 'author', 'doi',...)

{StBold}Output formatting:{Reset}
  {FgYellow}--fa --align-values{Reset}        pad fieldnames to align all values
  {FgYellow}--fc --comma-first{Reset}         comma first syntax (, title = ...)
  {FgYellow}--fl --no-trailing-comma{Reset}   don't add a last trailing comma
  {FgYellow}--fi --indent{Reset} {FgGreen}<space>{Reset}      space used for indentation, default is a tab
        Can be specified as an number (number of spaces) or a string with spaces
        and '_' 't' 'n' characters to mark space, tabs and newlines.

{StBold}Flags:{Reset}
  {FgYellow}-i --inplace{Reset}          Modify input files inplace
        ignores any specified output files
  {FgYellow}-f --force-overwrite{Reset}  Overwrite already present fields
        The default is to overwrite a field if it is empty or absent
  {FgYellow}-p --prefix{Reset}           Write new fields with a prefix
        eg: will write "BTACtitle = ..." instead of "title = ..." in the bib file.
        Can overwrite existing fields starting with BTACxxxx, even without the -f option.
        Can be combined with -f to safely show info for already present fields.

  {FgYellow}-t --timeout{Reset} {FgGreen}<float>{Reset}  set timeout on request, default: {TIMEOUT} s
        Set to -1 for no timeout.
  {FgYellow}-S --ignore-ssl{Reset}       Ignore SSL verification when performing queries

  {FgYellow}-d --dump-data{Reset} {FgGreen}<file.json>{Reset} writes all data from matching entries to
        the given file in JSON format, so data from multiple sources can be compared
  {FgYellow}-O --no-output{Reset}        Don't write any output files (except the --dump-data file)

  {FgYellow}-v --verbose{Reset}          increase verbosity (use up to 3 times)
  {FgYellow}-s --silent{Reset}           decrease verbosity (use up to 4 times)
  {FgYellow}-n --no-color{Reset}         don't color/stylise output

  {FgYellow}--version{Reset}             show version number
  {FgYellow}-h --help{Reset}             show this help

{StBold}Source and Bug reports:{Reset}
  The source code is available under an {LICENSE} License on github:
  {StUnderline}{URL}{Reset}

  You can report bugs there using the issue tracker."""



"""
Project-wide constants
"""

from typing import Dict

NAME = "bibtexautocomplete"
SCRIPT_NAME = "btac"
AUTHOR = "Dorian Lesbre"
URL = "https://github.com/dlesbre/bibtex-autocomplete"
LICENSE = "MIT"
DESCRIPTION = "Module to complete bibtex files by polling online databases"

VERSION_MAJOR = 1
VERSION_MINOR = 1
VERSION_PATCH = 6

VERSION_DATE = "2023-01-06"

VERSION = (VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH)
VERSION_STR = f"{VERSION_MAJOR}.{VERSION_MINOR}.{VERSION_PATCH}"

EMAIL = "dorian.lesbre" + chr(64) + "gmail.com"

# Minimum delay between queries to same host, to avoid surcharging server
MIN_QUERY_DELAY = 0.02  # s, so 50 per second
CONNECTION_TIMEOUT = 20.0  # seconds

USER_AGENT = f"{NAME}/{VERSION_STR} ({URL}; mailto:{EMAIL})"

EntryType = Dict[str, str]  # Type of a bibtex entry

MAX_THREAD_NB = 8  # Max number of threads

# Renaming pattern for nex files
# name is the filename, without extension ("ex" for "ex.bib")
# suffix is the file extansion, with leading dot (".bib" for "ex.bib")
BTAC_FILENAME = "{name}.btac{suffix}"

# Most APIs allow to limit the number of results returned for a search query
# This allows for smaller data transfers
QUERY_MAX_RESULTS = 10

# Prefix added to fields with -p / --prefix option
FIELD_PREFIX = "BTAC"

# Bullet printed to the screen when printing a list
BULLET = "{FgBlue}{StBold}*{Reset} "



"""
Lookup info from https://dlbp.org
"""

from typing import Dict, Iterable, List

from ..bibtex.author import Author
from ..bibtex.entry import BibtexEntry, FieldNames
from ..lookups.lookups import JSON_AT_Lookup
from ..utils.constants import QUERY_MAX_RESULTS
from ..utils.safe_json import SafeJSON


class DBLPLookup(JSON_AT_Lookup):
    """Lookup for info on https://dlbp.org
    Uses the API documented here:
    https://dblp.org/faq/13501473.html

    example URLs:
    no DOI mode.
    Author+title mode:
    https://dblp.org/search/publ/api?format=json&h=3&q=Lamiraux+Reactive+Path+Deformation+for+Nonholonomic+Mobile+Robots
    """

    name = "dblp"

    domain = "dblp.org"
    path = "/search/publ/api"

    def get_params(self) -> Dict[str, str]:
        search = ""
        if self.author is not None:
            search += self.author + " "
        if self.title is not None:
            search += self.title + " "
        return {"format": "json", "h": str(QUERY_MAX_RESULTS), "q": search.strip()}

    def get_results(self, data: bytes) -> Iterable[SafeJSON]:
        """Return the result list"""
        return SafeJSON.from_bytes(data)["result"]["hits"]["hit"].iter_list()

    @staticmethod
    def get_authors(info: SafeJSON) -> List[Author]:
        """Return a bibtex formatted list of authors"""
        authors = info["authors"]["author"]
        formatted = []
        for author in authors.iter_list():
            aut = Author.from_name(author["text"].to_str())
            if aut is not None:
                formatted.append(aut)
        return formatted

    def get_value(self, result: SafeJSON) -> BibtexEntry:
        info = result["info"]
        values = BibtexEntry()
        values.author = self.get_authors(info)
        values.doi = info["doi"].to_str()
        values.pages = info["pages"].to_str()
        values.title = info["title"].to_str()
        values.volume = info["volume"].to_str()
        values.url = info["ee"].to_str() if info["access"].to_str() == "open" else None
        values.year = info["year"].to_str()
        return values

    fields = {
        FieldNames.AUTHOR,
        FieldNames.DOI,
        FieldNames.PAGES,
        FieldNames.TITLE,
        FieldNames.VOLUME,
        FieldNames.URL,
        FieldNames.YEAR,
    }



"""
Wraps around bibtexparser's entry representations for safer access
and field normalization
"""

from typing import Any, Iterator, List, Optional, Set, Tuple

from ..utils.constants import EntryType
from .author import Author
from .normalize import get_field, has_data, has_field, normalize_doi, normalize_month


class FieldNames:
    """constants for bibtex field names"""

    ADDRESS = "address"
    ANNOTE = "annote"
    AUTHOR = "author"
    BOOKTITLE = "booktitle"
    CHAPTER = "chapter"
    DOI = "doi"
    EDITION = "edition"
    EDITOR = "editor"
    HOWPUBLISHED = "howpublished"
    INSTITUTION = "institution"
    ISSN = "issn"
    ISBN = "isbn"
    JOURNAL = "journal"
    MONTH = "month"
    NOTE = "note"
    NUMBER = "number"
    ORGANIZATION = "organization"
    PAGES = "pages"
    PUBLISHER = "publisher"
    SCHOOL = "school"
    SERIES = "series"
    TITLE = "title"
    TYPE = "type"
    URL = "url"
    VOLUME = "volume"
    YEAR = "year"


# Set of all fields
FieldNamesSet: Set[str] = {
    value
    for attr, value in vars(FieldNames).items()
    if isinstance(value, str) and "_" not in attr and attr.upper() == attr
}

# Fields actually searched for
SearchedFields = FieldNamesSet.copy()

# Set of fields with sanitized inputs
SpecialFields: Set[str] = {
    "author",
    "doi",
    "editor",
    "month",
}


class BibtexEntry:
    """A class to encapsulate bibtex entries
    Avoids spelling errors in field names
    and performs sanity checks

    Note: sanity check are performed when setting/getting attributes
    >>> entry.doi = "10.1234/123456"
    >>> entry.doi = "not_a_doi" # Will not do anything, fails quietly
    >>> entry.doi # "10.1234/123456"
    They are not performed on initialization !
    >>> entry = BibtexEntry({'doi': 'not_a_doi'})
    >>> entry.doi # None
    >>> entry.to_entry() # {'doi': 'not_a_doi'}

    Some fields have special treatment:
    - author and editor return/are set by a list of authors instead of a string
    - doi is formatted on get/set to remove leading url
    - month is formatted to "1" -- "12" if possible (recognizes "jan", "FeB.", "March"...)
    """

    address: Optional[str]
    annote: Optional[str]
    author: List[Author]
    booktitle: Optional[str]
    chapter: Optional[str]
    doi: Optional[str]
    edition: Optional[str]
    editor: List[Author]
    howpublished: Optional[str]
    institution: Optional[str]
    issn: Optional[str]
    isbn: Optional[str]
    journal: Optional[str]
    month: Optional[str]  # Number in "1" .. "12"
    note: Optional[str]
    number: Optional[str]
    organization: Optional[str]
    pages: Optional[str]
    publisher: Optional[str]
    school: Optional[str]
    series: Optional[str]
    title: Optional[str]
    type: Optional[str]
    url: Optional[str]
    volume: Optional[str]
    year: Optional[str]

    _entry: EntryType

    def __init__(self, entry: Optional[EntryType] = None):
        """Init is not thread safe"""
        if entry is None:
            self._entry = dict()
        else:
            self._entry = entry.copy()

    def __getattribute__(self, attr_name: str) -> Any:
        """Performs checks when returning Bibtex fields"""
        if attr_name in FieldNamesSet:
            if attr_name in SpecialFields:
                return super().__getattribute__("get_" + attr_name)()
            return get_field(self._entry, attr_name)
        return super().__getattribute__(attr_name)

    def __setattr__(self, attr_name: str, value: Any) -> None:
        """Performs checks when returning Bibtex fields"""
        if attr_name in FieldNamesSet:
            if attr_name in SpecialFields:
                super().__getattribute__("set_" + attr_name)(value)
                return None
            elif not has_data(value):
                # Delete attribute
                if attr_name in self._entry:
                    del self._entry[attr_name]
            else:
                self._entry[attr_name] = value
            return None
        super().__setattr__(attr_name, value)

    def __delattr__(self, attr_name: str) -> None:
        if attr_name in FieldNamesSet:
            if attr_name in self._entry:
                del self._entry[attr_name]
            return
        return super().__delattr__(attr_name)

    def __contains__(self, field: str) -> bool:
        """Is the given field non empty in this entry?"""
        return field in FieldNamesSet and has_field(self._entry, field)

    def __iter__(self) -> Iterator[Tuple[str, str]]:
        """Iterates through the fields of self"""
        return filter(
            lambda pair: pair[0] in FieldNamesSet and has_data(pair[1]),
            self._entry.items(),
        )

    def get_author(self) -> List[Author]:
        """Formats entry['author'] into Author list"""
        authors = get_field(self._entry, FieldNames.AUTHOR)
        if authors is not None:
            return Author.from_namelist(authors)
        return []

    def get_editor(self) -> List[Author]:
        """Formats entry['editor'] into Author list"""
        authors = get_field(self._entry, FieldNames.EDITOR)
        if authors is not None:
            return Author.from_namelist(authors)
        return []

    def set_author(self, authors: List[Author]) -> None:
        """set entry['author']"""
        if len(authors) == 0:
            if FieldNames.AUTHOR in self._entry:
                del self._entry[FieldNames.AUTHOR]
        else:
            self._entry[FieldNames.AUTHOR] = Author.list_to_bibtex(authors)

    def set_editor(self, authors: List[Author]) -> None:
        """set entry['editor']"""
        if len(authors) == 0:
            if FieldNames.EDITOR in self._entry:
                del self._entry[FieldNames.EDITOR]
        else:
            self._entry[FieldNames.EDITOR] = Author.list_to_bibtex(authors)

    def get_doi(self) -> Optional[str]:
        """get the doi without leading url"""
        return normalize_doi(get_field(self._entry, FieldNames.DOI))

    def set_doi(self, doi: str) -> None:
        """get the doi without leading url"""
        value = normalize_doi(doi)
        if value is None:
            if FieldNames.DOI in self._entry:
                del self._entry[FieldNames.DOI]
        else:
            self._entry[FieldNames.DOI] = value

    def get_month(self) -> Optional[str]:
        month = get_field(self._entry, FieldNames.MONTH)
        if month is None:
            return month
        return normalize_month(month)

    def set_month(self, month: Optional[str]) -> None:
        if month is not None:
            month = normalize_month(month)
            if month != "":
                self._entry[FieldNames.MONTH] = month
                return None
        if FieldNames.MONTH in self._entry:
            del self._entry[FieldNames.MONTH]



"""
Defining and configuring the logger
"""

import logging
from sys import stderr, stdout
from threading import current_thread, main_thread
from typing import Any

from .ansi import ansi_format
from .constants import NAME

Level = int


class LevelFilter(logging.Filter):
    def __init__(self, low: Level, high: Level):
        self._low = low
        self._high = high
        logging.Filter.__init__(self)

    def filter(self, record: logging.LogRecord) -> bool:
        if self._low <= record.levelno <= self._high:
            return True
        return False


# custom level
VERBOSE_INFO = logging.INFO - 2
VERBOSE_DEBUG = logging.DEBUG - 2
VERY_VERBOSE_DEBUG = VERBOSE_DEBUG - 2
logging.addLevelName(VERBOSE_INFO, "VERBOSE_INFO")
logging.addLevelName(VERBOSE_DEBUG, "VERBOSE_DEBUG")
logging.addLevelName(VERY_VERBOSE_DEBUG, "VERY_VERBOSE_DEBUG")

DEFAULT_LEVEL = logging.INFO


class Logger:

    logger: logging.Logger

    def __init__(self, logger_name: str) -> None:
        # create logger
        self.logger = logging.getLogger(logger_name)
        # Errors got to STDERR
        self.error_handler = logging.StreamHandler(stderr)
        self.error_handler.addFilter(LevelFilter(logging.WARN, logging.CRITICAL))
        self.error_handler.setFormatter(logging.Formatter("%(message)s"))
        self.logger.addHandler(self.error_handler)
        # Everything else goes to STDOUT
        self.info_handler = logging.StreamHandler(stdout)
        self.info_handler.addFilter(LevelFilter(0, logging.WARN - 1))
        self.info_handler.setFormatter(logging.Formatter("%(message)s"))
        self.logger.addHandler(self.info_handler)

        self.logger.setLevel(DEFAULT_LEVEL)

    @staticmethod
    def add_thread_info(message: str) -> str:
        """Add thread name to message if not in main thread"""
        current = current_thread()
        if current is not main_thread():
            info = ansi_format("[{FgBlue}" + current.name + "{Reset}] ")
            len_info = len(current.name) + 3
            message = info + message.replace("\n", "\n" + " " * len_info)
        return message

    def to_logger(self, level: int, message: str, *args: Any, **kwargs: Any) -> None:
        """Formats a message (with given args and ansi colors)
        and sends it to the logger with the given level"""
        message = self.add_thread_info(ansi_format(message, *args, **kwargs))
        self.logger.log(level=level, msg=message)

    def warn(
        self, message: str, error: str = "WARNING", *args: Any, **kwargs: Any
    ) -> None:
        """Issue a warning, extra arguments are formatter options"""
        self.to_logger(
            logging.WARN,
            "{FgPurple}{error}:{Reset} " + message,
            *args,
            error=error,
            **kwargs
        )

    def error(
        self, message: str, error: str = "ERROR", *args: Any, **kwargs: Any
    ) -> None:
        """Issue an error, extra arguments are formatter options"""
        self.to_logger(
            logging.ERROR,
            "{FgRed}{error}:{Reset} " + message,
            error=error,
            *args,
            **kwargs
        )

    def critical(
        self, message: str, error: str = "CRITICAL ERROR", *args: Any, **kwargs: Any
    ) -> None:
        """Issue a critical error, extra arguments are formatter options"""
        self.to_logger(
            logging.CRITICAL,
            "{FgRed}{error}:{Reset} " + message,
            *args,
            error=error,
            **kwargs
        )

    def info(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Show info, extra arguments are formatter options"""
        self.to_logger(logging.INFO, message, *args, **kwargs)

    def verbose_info(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Show info when verbose, extra arguments are formatter options"""
        self.to_logger(VERBOSE_INFO, message, *args, **kwargs)

    def debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Show debug info, extra arguments are formatter options"""
        self.to_logger(logging.DEBUG, message, *args, **kwargs)

    def verbose_debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Show verbose debug info, extra arguments are formatter options"""
        self.to_logger(VERBOSE_DEBUG, message, *args, **kwargs)

    def very_verbose_debug(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Show very verbose debug info, extra arguments are formatter options"""
        self.to_logger(VERY_VERBOSE_DEBUG, message, *args, **kwargs)

    def set_level(self, level: Level) -> None:
        """Set the logger's level, using logging's level values"""
        self.logger.setLevel(level)

    verbosity = {
        -4: logging.CRITICAL + 10,
        -3: logging.CRITICAL,
        -2: logging.ERROR,
        -1: logging.WARN,
        0: logging.INFO,
        1: VERBOSE_INFO,
        2: logging.DEBUG,
        3: VERBOSE_DEBUG,
        4: VERY_VERBOSE_DEBUG,
    }

    def set_verbosity(self, verbosity: int) -> None:
        """Set verbosity:
        0 is default
        1, 2, 3 show more and more info
        -1, -2, -3, -4 show less and less (warning, error, critical errors, then nothing)"""
        maxi = max(self.verbosity)
        mini = min(self.verbosity)
        if verbosity > maxi:
            verbosity = maxi
        if verbosity < mini:
            verbosity = mini
        self.set_level(self.verbosity[verbosity])

    def header(self, title: str, level: Level = logging.INFO) -> None:
        """Shows a pretty header, 100% inspired by opam's output"""
        self.to_logger(level, "")  # newline
        title = (
            "{FgBlue}===={Reset} {StBold}"
            + title
            + "{Reset} {FgBlue}"
            + ("=" * (74 - len(title)))
            + "{Reset}"
        )
        self.to_logger(level, title)


logger = Logger(NAME)


class Hint:
    """Hints are messages displayed only once"""

    prefix = "{FgBlue}Hint:{Reset} "
    indent = " " * 6

    emitted: bool
    message: str

    def __init__(self, message: str) -> None:
        self.message = self.prefix + message.strip().replace("\n", self.indent + "\n")
        self.emitted = False

    def emit(self) -> None:
        if not self.emitted:
            self.emitted = True
            logger.info(self.message)



from typing import List

from ..lookups.abstract_base import LookupType
from .arxiv import ArxivLookup
from .crossref import CrossrefLookup
from .dblp import DBLPLookup
from .researchr import ResearchrLookup
from .unpaywall import UnpaywallLookup

# List of lookup to use, in the order they will be used
LOOKUPS: List[LookupType] = [
    CrossrefLookup,
    ArxivLookup,
    DBLPLookup,
    ResearchrLookup,
    UnpaywallLookup,
]
LOOKUP_NAMES = [cls.name for cls in LOOKUPS]

__all__ = (
    "LookupType",
    "LOOKUPS",
    "LOOKUP_NAMES",
    "CrossrefLookup",
    "ArxivLookup",
    "DBLPLookup",
    "ResearchrLookup",
    "UnpaywallLookup",
)



"""
Bibtexautocomplete
main class used to manage calls to different lookups
"""

from functools import reduce
from json import dump as json_dump
from pathlib import Path
from threading import Condition
from typing import (
    Callable,
    Container,
    Dict,
    Iterable,
    Iterator,
    List,
    Tuple,
    TypeVar,
    cast,
)

from alive_progress import alive_bar  # type: ignore
from bibtexparser.bibdatabase import BibDatabase

from ..APIs.doi import DOICheck, URLCheck
from ..bibtex.entry import BibtexEntry, FieldNames
from ..bibtex.io import file_read, file_write, get_entries
from ..bibtex.normalize import has_field
from ..lookups.abstract_base import LookupType
from ..utils.constants import BULLET, FIELD_PREFIX, MAX_THREAD_NB, EntryType
from ..utils.logger import VERBOSE_INFO, logger
from ..utils.only_exclude import OnlyExclude
from .data_dump import DataDump
from .threads import LookupThread

T = TypeVar("T")
Q = TypeVar("Q")


def memoize(method: Callable[[T], Q]) -> Callable[[T], Q]:
    """Simple decorator for no argument method memoization
    as an attribute"""
    attribute = "_memoize_" + method.__name__

    def new_method(self: T) -> Q:
        if not hasattr(self, attribute):
            setattr(self, attribute, method(self))
        return cast(Q, getattr(self, attribute))

    return new_method


class BibtexAutocomplete(Iterable[EntryType]):
    """Main class used to dispatch calls to the relevant lookups"""

    bibdatabases: List[BibDatabase]
    lookups: List[LookupType]
    fields: Container[str]
    entries: OnlyExclude[str]
    force_overwrite: bool
    prefix: str
    dumps: List[DataDump]

    changed_fields: int
    changed_entries: int

    # Ordered list of (entry, changes) where
    # changes is a list of (field, new_value, source)
    changes: List[Tuple[str, List[Tuple[str, str, str]]]]

    def __init__(
        self,
        bibdatabases: List[BibDatabase],
        lookups: Iterable[LookupType],
        fields: Container[str],
        entries: OnlyExclude[str],
        force_overwrite: bool,
        prefix: bool = False,
    ):
        self.bibdatabases = bibdatabases
        self.lookups = list(lookups)
        self.fields = fields
        self.entries = entries
        self.force_overwrite = force_overwrite
        self.changed_entries = 0
        self.changed_fields = 0
        self.changes = []
        self.dumps = []
        self.prefix = FIELD_PREFIX if prefix else ""

    def __iter__(self) -> Iterator[EntryType]:
        """Iterate through entries"""
        for db in self.bibdatabases:
            yield from filter(lambda x: x["ID"] in self.entries, get_entries(db))

    @memoize
    def count_entries(self) -> int:
        """count the number of entries"""
        # Its official, functional programming has infected me...
        return reduce(lambda x, _y: x + 1, self, 0)

    def print_filters(self) -> None:
        """Prints entry filter effects"""
        all_entries = [x["ID"] for db in self.bibdatabases for x in get_entries(db)]
        total = len(all_entries)
        filtered = self.count_entries()
        if total > filtered:
            logger.info("Filtered down to {} entries".format(filtered))
        warn_only, warn_exclude = self.entries.unused(all_entries)
        for x in sorted(warn_only):
            logger.warn('No entry with ID "{ID}"', ID=x)
        for x in sorted(warn_exclude):
            logger.warn('No entry with ID "{ID}"', ID=x)

    @memoize
    def get_id_padding(self) -> int:
        """Return the max length of entries' ID
        to use for pretty printing"""
        max_id_padding = 40
        return min(
            max((len(entry["ID"]) + 1 for entry in self), default=0), max_id_padding
        )

    def autocomplete(self, no_progressbar: bool = False) -> None:
        """Main function that does all the work
        Iterate through entries, performing all lookups"""
        logger.header("Completing entries")
        total = self.count_entries() * len(self.lookups)
        entries = list(self)
        condition = Condition()
        assert len(self.lookups) < MAX_THREAD_NB
        threads: List[LookupThread] = []
        with alive_bar(
            total,
            title="Querying databases:",
            disable=no_progressbar,
            enrich_print=False,
            monitor="[{percent:.0%}]",
            monitor_end="[{percent:.0%}]",
            stats="(eta: {eta})",
            stats_end="",
            dual_line=True,
        ) as bar:
            # Create all threads
            for lookup in self.lookups:
                threads.append(LookupThread(lookup, entries, condition, bar))
            condition.acquire()
            # Start all threads
            for thread in threads:
                thread.start()
            position = 0
            nb_entries = self.count_entries()
            while position < nb_entries:
                # Check if all threads have resolved the current entry
                step = True
                thread_positions = []
                for thread in threads:
                    if position >= thread.position:
                        step = False
                    thread_positions.append(
                        f"[{thread.lookup.name}: {thread.position}/{nb_entries}]"
                    )
                bar.text = (
                    " ".join(thread_positions)
                    + f" Found {self.changed_fields} new fields"
                )
                if not step:  # Some threads have not found data for current entry
                    condition.wait()
                else:  # update data for current entry
                    self.update_entry(entries[position], threads, position)
                    position += 1
        logger.info(
            "Modified {changed_entries} / {count_entries} entries"
            ", added {changed_fields} fields",
            changed_entries=self.changed_entries,
            count_entries=self.count_entries(),
            changed_fields=self.changed_fields,
        )

    def update_entry(
        self, entry: EntryType, threads: List[LookupThread], position: int
    ) -> None:
        """Reads all data the threads have found on a new entry,
        and uses it to update the entry with new fields"""
        changes: List[Tuple[str, str, str]] = []
        dump = DataDump(entry["ID"])
        iterator = reversed(threads) if self.force_overwrite else threads
        for thread in iterator:
            result, info = thread.result[position]
            dump.add_entry(thread.lookup.name, result, info)
            if result is not None:
                to_add = self.combine(entry, result)
                to_add = self.sanitize(to_add)
                entry.update({self.prefix + field: to_add[field] for field in to_add})
                changes.extend(
                    (field, value, thread.name) for field, value in to_add.items()
                )
        dump.new_fields = len(changes)
        if changes != []:
            self.changed_entries += 1
            self.changed_fields += len(changes)
            self.changes.append((entry["ID"], changes))
        logger.verbose_info(
            BULLET + "{StBold}{entry}{StBoldOff} {nb} new fields",
            entry=entry["ID"].ljust(self.get_id_padding()),
            nb=len(changes),
        )
        self.dumps.append(dump)

    def combine(self, entry: EntryType, new_info: BibtexEntry) -> Dict[str, str]:
        """Adds the information in info to entry.
        Does not overwrite unless self.force_overwrite is True
        only acts on fields contained in self.fields"""
        changes: Dict[str, str] = {}
        for field, value in new_info:
            if field not in self.fields:
                continue
            # Does the field actually contain any value
            if value is None:
                continue
            s_value = str(value)
            if s_value.strip() == "":
                continue
            # Is it present on entry
            if self.force_overwrite or (not has_field(entry, field)):
                changes[field] = s_value
        return changes

    def sanitize(self, changes: Dict[str, str]) -> Dict[str, str]:
        """Runs sanity checks on new_info if needed"""
        doi = changes.get(FieldNames.DOI)
        url = changes.get(FieldNames.URL)
        if doi is not None:
            doi_checker = DOICheck(doi)
            if doi_checker.query() is not True:
                del changes[FieldNames.DOI]
        if url is not None:
            checker = URLCheck(url)
            if checker.query() is None:
                del changes[FieldNames.URL]
        return changes

    def print_changes(self) -> None:
        """prints a pretty list of changes"""
        logger.header("New fields", VERBOSE_INFO)
        if self.changes == []:
            logger.verbose_info("No new fields")
            return None
        for entry, changes in sorted(self.changes, key=lambda x: x[0]):
            logger.verbose_info(
                BULLET + "{StBold}{entry}{Reset}:",
                entry=entry,
            )
            for field, value, source in changes:
                logger.verbose_info(
                    "    {FgBlue}{field}{Reset}{FgWhite}{StFaint} = {{{Reset}"
                    "{value}{FgWhite}{StFaint}}},{Reset}"
                    " {FgGreen}{StItalics}% {source}{Reset}",
                    field=field,
                    value=value,
                    source=source,
                )

    wrote_header = False

    def write_header(self) -> None:
        """Ensures the final section header is only written once"""
        if not self.wrote_header:
            logger.header("Writing files")
        self.wrote_header = True

    def write_dumps(self, path: Path) -> None:
        """Write the dumps to the given file"""
        self.write_header()
        json = [dump.to_dict() for dump in self.dumps]
        logger.info("Writing data dump to '{}'", str(path))
        try:
            with open(path, "w", encoding="utf-8") as file:
                json_dump(json, file, indent=2)
        except (IOError, UnicodeDecodeError) as err:
            logger.error(
                "Failed to dump data to '{path}' : {FgPurple}{err}{Reset}",
                path=str(path),
                err=err,
            )

    def write(self, files: List[Path]) -> None:
        """Writes the databases in self to the given files
        If not enough files, extra databases are written to stdout
        If too many files, extras are ignored"""
        self.write_header()
        total = len(self.bibdatabases)
        wrote = 0
        for i, db in enumerate(self.bibdatabases):
            file = files[i]
            logger.info(
                "Writing file {id} / {total} to '{file}'",
                id=i + 1,
                total=total,
                file=file,
            )
            wrote += file_write(file, db)
        logger.info(
            "Wrote {total} {files}",
            total=wrote,
            files="file" if wrote == 1 else "files",
        )

    @staticmethod
    def read(files: List[Path]) -> List[BibDatabase]:
        logger.header("Reading files")
        length = len(files)
        dbs = []
        for i, file in enumerate(files):
            logger.info(
                "Reading file {id} / {length} from '{file}'",
                id=i + 1,
                length=length,
                file=file,
            )
            dbs.append(file_read(file))
        nb_entries = sum(len(get_entries(db)) for db in dbs)
        logger.info(
            "Read {nb_entries} {entry} from {total} {file}",
            total=length,
            nb_entries=nb_entries,
            entry="entries" if nb_entries != 1 else "entry",
            file="files" if length != 1 else "file",
        )
        return dbs



"""
Lookup info from https://unpaywall.org/
"""

from typing import Dict, Iterable, List, Optional
from urllib.parse import urlencode

from ..bibtex.author import Author
from ..bibtex.entry import BibtexEntry, FieldNames
from ..lookups.lookups import JSON_DT_Lookup
from ..utils.constants import EMAIL
from ..utils.safe_json import SafeJSON


class UnpaywallLookup(JSON_DT_Lookup):
    """Lookup on https://unpaywall.org/
    only if the entry has a known DOI
    API documented at:
    https://unpaywall.org/products/api

    Example urls:
    DOI mode:
    https://api.unpaywall.org/v2/10.1109/tro.2004.829459?email=some.test%40gmail.com

    Title mode:
    https://api.unpaywall.org/v2/search/?email=some.test%40gmail.com&query=Reactive+Path+Deformation+for+Nonholonomic+Mobile+Robots
    """

    name = "unpaywall"

    domain = "api.unpaywall.org"
    path = "/v2/"

    params = {"email": EMAIL}

    doi: Optional[str] = None
    title: Optional[str] = None

    def get_params(self) -> Dict[str, str]:
        base = super().get_params()
        if self.doi is None:
            if self.title is None:
                raise ValueError("query with no title or doi")
            base["query"] = self.title
        return base

    def get_path(self) -> str:
        base = self.path
        params = "?" + urlencode(self.get_params())
        if self.doi is not None:
            return base + self.doi + params
        return base + "search/" + params

    def get_results(self, data: bytes) -> Optional[Iterable[SafeJSON]]:
        json = SafeJSON.from_bytes(data)
        if self.doi is not None:
            # doi based search, single result if any
            return [json]
        return json["results"].iter_list()

    @staticmethod
    def get_authors(authors: SafeJSON) -> List[Author]:
        """Return a bibtex formatted list of authors"""
        formatted = []
        for author in authors.iter_list():
            family = author["family"].to_str()
            if family is not None:
                given = author["given"].to_str()
                formatted.append(Author(family, given))
        return formatted

    def get_value(self, result: SafeJSON) -> BibtexEntry:
        result = result["response"]
        date = result["published_date"].to_str()  # ISO format YYYY-MM-DD
        year = str(result["year"].to_int())
        month = None
        if date is not None:
            if year is None and len(date) >= 4:
                year = date[0:4]
            # Unknown month are set to first january...
            if date[4:] != "-01-01" and len(date) >= 7:
                month = date[5:7]
        values = BibtexEntry()

        title = result["journal_name"].to_str()
        is_journal = result["genre"].to_str() == "journal-article"

        values.author = self.get_authors(result["z_authors"])
        values.booktitle = None if is_journal else title
        values.doi = result["doi"].to_str()
        values.issn = result["journal_issn_l"].to_str()
        values.journal = title if is_journal else None
        values.month = month
        values.publisher = result["publisher"].to_str()
        values.title = result["title"].to_str()
        values.url = result["best_oa_location"]["url_for_pdf"].to_str()
        values.year = year

        return values

    fields = {
        FieldNames.AUTHOR,
        FieldNames.BOOKTITLE,
        FieldNames.DOI,
        FieldNames.ISSN,
        FieldNames.JOURNAL,
        FieldNames.MONTH,
        FieldNames.PUBLISHER,
        FieldNames.TITLE,
        FieldNames.URL,
        FieldNames.YEAR,
    }



"""
Abstract base classes used to represent queries

LookupProtocol(Protocol): duck-typing of lookups. They must have
  - an attribute name : str (UNIQUE)
  - a method query(self) -> Optional[BibtexEntry]
  - a method __init__(self, entry: Entry)
  - a method get_last_query_info(self) -> Dict[str, JSONType]
        with return extra information to add to data-dumps


AbstractLookup(): abstract base class, defines abstract methods

AbstractEntryLookup(AbstractLookup): entry attribute and __init__

AbstractDataLookup(AbstractLookup): split query into two new methods:
  - get_data : Self -> Data - to perform the actual query
  - process_data : Self, Data -> BibtexEntry - process data into a bibtex entry
"""

from typing import (
    ClassVar,
    Dict,
    Generic,
    NamedTuple,
    Optional,
    Protocol,
    Type,
    TypeVar,
)

from ..bibtex.entry import BibtexEntry
from ..utils.safe_json import JSONType

Input = TypeVar("Input", covariant=True)
Output = TypeVar("Output", covariant=True)


class LookupProtocol(Protocol, Generic[Input, Output]):
    name: ClassVar[str]  # used to identify the lookup, also appears in help string

    def query(self) -> Optional[Output]:
        """Performs one or more queries to try and obtain the result
        VIRTUAL METHOD : must be overridden"""
        raise NotImplementedError("should be overridden in child class")

    def get_last_query_info(self) -> Dict[str, JSONType]:
        """Extra information to add to the data-dump about this query"""
        return dict()

    def __init__(self, input: Input) -> None:
        pass


LookupType = Type[LookupProtocol[BibtexEntry, BibtexEntry]]


class AbstractLookup(Generic[Input, Output]):
    """Abstract base class for lookup
    Is a valid LookupProtocol, but all methods should be overridden

    Use this as a base class for mixin so that super().query calls can typecheck"""

    name: ClassVar[str]  # used to identify the lookup, also appears in help string

    def query(self) -> Optional[Output]:
        """Performs one or more queries to try and obtain the result
        VIRTUAL METHOD : must be overridden"""
        raise NotImplementedError("should be overridden in child class")

    def get_last_query_info(self) -> Dict[str, JSONType]:
        """Extra information to add to the data-dump about this query"""
        return dict()

    def __init__(self, input: Input) -> None:
        pass


class AbstractEntryLookup(AbstractLookup[BibtexEntry, BibtexEntry]):
    """Abstract minimal lookup,
    Implements simple __init__ putting the argument in self.entry

    Virtual methods and attributes : (must be overridden in children):
    - name : str
    - query: Self -> Optional[BibtexEntry]
    """

    entry: BibtexEntry

    def __init__(self, input: BibtexEntry) -> None:
        super().__init__(input)
        self.entry = input


class Data(NamedTuple):
    data: bytes
    code: int
    reason: str
    delay: float


class AbstractDataLookup(AbstractLookup[Input, Output]):
    def get_data(self) -> Optional[Data]:
        """Performs a query to get data from the server
        VIRTUAL METHOD : must be overridden"""
        raise NotImplementedError("should be overridden in child class")

    def process_data(self, data: Data) -> Optional[Output]:
        """Should create a new entry with info extracted from data
        VIRTUAL METHOD : must be overridden"""
        raise NotImplementedError("should be overridden in child class")

    def query(self) -> Optional[Output]:
        """Tries to complete an entry
        override this to make multiple requests
        (i.e. try different search terms)"""
        data = self.get_data()
        if data is not None:
            return self.process_data(data)
        return None



"""
setup file. run 'python3 setup.py install' to install.
"""
from setuptools import find_packages, setup  # type: ignore

import bibtexautocomplete

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name=bibtexautocomplete.__name__,
    version=bibtexautocomplete.__version__,
    author=bibtexautocomplete.__author__,
    author_email=bibtexautocomplete.__email__,
    url=bibtexautocomplete.__url__,
    description=bibtexautocomplete.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    entry_points={
        "console_scripts": "btac = bibtexautocomplete.core:main",
    },
    install_requires=[
        "bibtexparser",
        "alive-progress>=3.0.0",
    ],
    extras_require={
        "dev": [
            "pre-commit",
            "pytest",
            "mypy",
            "black",
            "flake8",
            "isort",
            "coverage",
            "pytest-cov",
        ],
        "deploy": ["wheel", "twine"],
    },
    python_requires=">=3.8",
    license="MIT",
    platforms=["any"],
    keywords=["bibtex biblatex latex autocomplete btac"],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "Natural Language :: English",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Markup :: LaTeX",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Typing :: Typed",
    ],
    data_files=[("", ["LICENSE"])],
)



from typing import Any, Dict, Optional

from ..bibtex.entry import BibtexEntry
from ..utils.safe_json import JSONType


class DataDump:
    """Contains fields for all entries"""

    results: Dict[str, Optional[Dict[str, JSONType]]]
    id: str
    new_fields: int

    def __init__(self, id: str) -> None:
        self.id = id
        self.new_fields = 0
        self.results = {}

    def add_entry(
        self, lookup_name: str, entry: Optional[BibtexEntry], info: Dict[str, JSONType]
    ) -> None:
        if entry is None:
            self.results[lookup_name] = None
            return
        infos = {"query-" + key: val for key, val in info.items()}
        infos.update({key: val for key, val in entry._entry.items() if val is not None})
        self.results[lookup_name] = infos

    def to_dict(self) -> Dict[str, Any]:
        json: Dict[str, Any] = {
            "entry": self.id,
            "new-fields": self.new_fields,
        }
        json.update(self.results)
        return json



"""
ANSI escape sequence for colors and styles
"""

from typing import Any


class ANSICodes:

    use_ansi: bool = True

    Codes = {
        # Foreground Colors
        "FgBlack": "\x1b[30m",
        "FgRed": "\x1b[31m",
        "FgGreen": "\x1b[32m",
        "FgYellow": "\x1b[33m",
        "FgBlue": "\x1b[34m",
        "FgPurple": "\x1b[35m",
        "FgCyan": "\x1b[36m",
        "FgWhite": "\x1b[37m",
        # Foreground Highlight Colors
        "FgHBlack": "\x1b[90m",
        "FgHRed": "\x1b[91m",
        "FgHGreen": "\x1b[92m",
        "FgHYellow": "\x1b[93m",
        "FgHBlue": "\x1b[94m",
        "FgHPurple": "\x1b[95m",
        "FgHCyan": "\x1b[96m",
        "FgHWhite": "\x1b[97m",
        "FgReset": "\x1b[38m",
        # Background Colors
        "BgBlack": "\x1b[40m",
        "BgRed": "\x1b[41m",
        "BgGreen": "\x1b[42m",
        "BgYellow": "\x1b[43m",
        "BgBlue": "\x1b[44m",
        "BgPurple": "\x1b[45m",
        "BgCyan": "\x1b[46m",
        "BgWhite": "\x1b[47m",
        # Background Highlight Colors
        "BgHBlack": "\x1b[100m",
        "BgHRed": "\x1b[101m",
        "BgHGreen": "\x1b[102m",
        "BgHYellow": "\x1b[103m",
        "BgHBlue": "\x1b[104m",
        "BgHPurple": "\x1b[105m",
        "BgHCyan": "\x1b[106m",
        "BgHWhite": "\x1b[107m",
        "BgReset": "\x1b[48m",
        # Styles
        "StBold": "\x1b[1m",
        "StBoldOff": "\x1b[22m",
        "StFaint": "\x1b[2m",
        "StFaintOff": "\x1b[22m",
        "StItalics": "\x1b[3m",
        "StItalicsOff": "\x1b[23m",
        "StUnderline": "\x1b[4m",
        "StUnderlineOff": "\x1b[24m",
        "StBlink": "\x1b[5m",
        "StBlinkOff": "\x1b[25m",
        "StInverse": "\x1b[7m",
        "StInverseOff": "\x1b[0m",
        "StBarred": "\x1b[9m",
        "StBarredOff": "\x1b[29m",
        "StOverline": "\x1b[53m",
        "StOverlineOff": "\x1b[55m",
        # Extras
        "Reset": "\x1b[0m",
        "Clearscreen": "\x1bc",
        "Clearline": "\x1b[2K\x1b[1G",
    }

    # Empty dict used when disabled
    EmptyCodes = {attr: "" for attr in Codes}

    @classmethod
    def ansi_format(cls, string: str, *args: Any, **kwargs: Any) -> str:
        """Return the string formatted with args and kwargs,
        adding the color formatters"""
        codes = cls.Codes
        if not cls.use_ansi:
            codes = cls.EmptyCodes
        return string.format(*args, **kwargs, **codes)


ansi_format = ANSICodes.ansi_format



"""
Lookup info from https://www.crossref.org
"""

from typing import Dict, Iterable, List, Optional, Tuple

from ..bibtex.author import Author
from ..bibtex.entry import BibtexEntry, FieldNames
from ..lookups.lookups import JSON_DAT_Lookup
from ..utils.constants import QUERY_MAX_RESULTS
from ..utils.safe_json import SafeJSON


class CrossrefLookup(JSON_DAT_Lookup):
    """Lookup info on https://www.crossref.org
    Uses the crossref REST API, documentated here:
    https://api.crossref.org/swagger-ui/index.html

    example URLs:
    DOI mode:
    https://api.crossref.org/works/10.1109/tro.2004.829459
    Author + title:
    https://api.crossref.org/works?rows=3&query.title=Reactive+Path+Deformation+for+Nonholonomic+Mobile+Robots&query.author=Lamiraux
    """

    name = "crossref"

    domain = "api.crossref.org"
    path = "/works"

    single_author_queries = False  # Crossref is a bit slow, don't do too many queries

    def get_path(self) -> str:
        if self.doi is not None:
            return self.path + "/" + self.doi
        return super().get_path()

    def get_params(self) -> Dict[str, str]:
        base = {"rows": str(QUERY_MAX_RESULTS)}
        if self.title is not None:
            base["query.title"] = self.title
        if self.author is not None:
            base["query.author"] = self.author
        return base

    def update_rate_cap(self) -> Optional[float]:
        if self.response is None:
            return None
        limit = self.response.getheader("X-Rate-Limit-Limit")
        interval = self.response.getheader("X-Rate-Limit-Interval")
        if limit is None or interval is None:
            return None
        try:
            return float(interval[:-1]) / float(limit)
        except (ValueError, ZeroDivisionError):
            return None

    def get_results(self, data: bytes) -> Optional[Iterable[SafeJSON]]:
        """Return the result list"""
        json = SafeJSON.from_bytes(data)
        if json["status"].to_str() == "ok":
            message = json["message"]
            if self.doi is not None:
                return [message]
            return message["items"].iter_list()
        return None

    @staticmethod
    def get_authors(authors: SafeJSON) -> List[Author]:
        """Parses JSON output into bibtex formatted author list"""
        formatted = []
        for author in authors.iter_list():
            lastname = author["family"].to_str()
            if lastname is not None:
                formatted.append(Author(lastname, author["given"].to_str()))
        return formatted

    @staticmethod
    def get_date(result: SafeJSON) -> Tuple[Optional[str], Optional[str]]:
        date = None
        for field in (
            "published-print",
            "issued",
            "published-online",
            "created",
            "content-created",
        ):
            date = result[field]["date-parts"][0]
            year = date[0].force_str()
            month = date[1].force_str()
            if year is not None:
                return year, month
        return None, None

    def get_value(self, result: SafeJSON) -> BibtexEntry:
        """Extract bibtex data from JSON output"""
        year, month = self.get_date(result)
        title = result["container-title"][0].to_str()
        is_journal = result["type"].to_str() == "journal-article"
        values = BibtexEntry()
        values.author = self.get_authors(result["author"])
        values.booktitle = None if is_journal else title
        values.doi = result["DOI"].to_str()
        values.issn = result["ISSN"][0].to_str()
        values.isbn = result["ISBN"][0].to_str()
        values.journal = title if is_journal else None
        values.month = month
        values.pages = result["page"].to_str()
        values.publisher = result["publisher"].to_str()
        values.title = result["title"][0].to_str()
        values.volume = result["volume"].to_str()
        values.year = year
        return values

    fields = {
        FieldNames.AUTHOR,
        FieldNames.BOOKTITLE,
        FieldNames.DOI,
        FieldNames.ISSN,
        FieldNames.ISBN,
        FieldNames.JOURNAL,
        FieldNames.MONTH,
        FieldNames.PAGES,
        FieldNames.PUBLISHER,
        FieldNames.TITLE,
        FieldNames.VOLUME,
        FieldNames.YEAR,
    }



"""
Mixin to perform multiple queries according to entry information
Set class attribute to be used by the query in get_data
"""

from typing import Iterator, Optional

from ..bibtex.entry import BibtexEntry
from ..bibtex.normalize import normalize_str
from .abstract_base import AbstractEntryLookup, AbstractLookup, Input, Output


class MultipleQueryMixin(AbstractLookup[Input, Output]):
    """Mixin to perform multiple queries

    Defines:
    - iter_queries : Self -> Iterator[None] - empty iterator (should be overridden)
    - query : Self -> Optional[BibtexEntry] - call parent query as long as iter_query yields
        stop a first valid (non None) value
    """

    def iter_queries(self) -> Iterator[None]:
        """Used to iterate through the queries
        The yielded result doesn't really matter.
        This method should dynamically modify attributes (like self.title)
        which are then used when constructing queries (i.e. in get_param())
        Default behavior: no queries"""
        return iter([])

    def query(self) -> Optional[Output]:
        """Performs queries as long as iter_query yields
        Stops at first valid result found"""
        for _ in self.iter_queries():
            value = super().query()
            if value is not None:
                return value
        return None


class DOIQueryMixin(MultipleQueryMixin[BibtexEntry, BibtexEntry], AbstractEntryLookup):
    """Performs one query setting self.doi if self.entry has a doi
    then parent queries if any
    """

    doi: Optional[str] = None

    def iter_queries(self) -> Iterator[None]:
        self.doi = self.entry.doi
        if self.doi is not None:
            yield None
        # Perform more queries without doi
        self.doi = None
        for x in super().iter_queries():
            yield x


class TitleQueryMixin(
    MultipleQueryMixin[BibtexEntry, BibtexEntry], AbstractEntryLookup
):
    """Sets self.title if self.entry has a title
    Performs parent queries if any
    then perform a single query if self.title is set
    """

    title: Optional[str] = None

    def iter_queries(self) -> Iterator[None]:
        self.title = self.entry.title
        # Perform parent queries with title set
        for x in super().iter_queries():
            yield x
        if self.title is not None:
            yield None


class TitleAuthorQueryMixin(
    MultipleQueryMixin[BibtexEntry, BibtexEntry], AbstractEntryLookup
):
    """
    Sets self.title to entry title if any
    Sets self.author if self.entry has a authors to space separated list of lastnames
    Performs parent queries if any
    if self.title:
      performs a single query if self.author is not None
      performs a query per author
        (resetting self.author to just one author)
        if more than one author and self.single_author_queries is set
      unset self.author
      performs a single query
    """

    entry: BibtexEntry
    author_join: str = " "
    single_author_queries: bool = True
    title: Optional[str] = None
    author: Optional[str] = None

    max_author_queries: int = 10

    def iter_queries(self) -> Iterator[None]:
        # Find and format authors
        self.title = (
            None if self.entry.title is None else normalize_str(self.entry.title)
        )
        authors = self.entry.author
        if authors:
            self.author = self.author_join.join(author.lastname for author in authors)
        # Perform parent queries
        for x in super().iter_queries():
            yield x
        if self.title is None:
            return None
        # Perform one query with all authors
        yield None
        if len(authors) > 1 and self.single_author_queries:
            for ii, author in enumerate(authors):
                # Perform one query per author, with at most 10 queries
                if ii > self.max_author_queries:
                    break
                self.author = author.lastname
                yield None
        self.author = None
        if authors:
            yield None


class DATQueryMixin(TitleAuthorQueryMixin, DOIQueryMixin):
    """DOI - Authors - Title Mixin
    queries in order:
    - if doi, with doi, with title (if any), with all authors (if any)
    - if authors, with all authors, with title (if any)
    - if authors, with each author individually, with title (if any)
    - if title, with title (if any)"""

    pass


class DTQueryMixin(TitleAuthorQueryMixin, DOIQueryMixin):
    """DOI - Title Mixin
    queries in order:
    - if doi, with doi, with title (if any)
    - if title, with title (if any)"""

    pass
